@@ -1,6 +1,6 @@
 /***********************************************************************
  *
- * Copyright (C) 2009, 2010 Graeme Gott <graeme@gottcode.org>
+ * Copyright (C) 2009, 2010, 2011 Graeme Gott <graeme@gottcode.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,12 +19,16 @@
 
 #include "board.h"
 
+#include "beveled_rect.h"
 #include "clock.h"
 #include "generator.h"
+#include "language_settings.h"
 #include "letter.h"
+#include "random.h"
 #include "scores_dialog.h"
 #include "solver.h"
 #include "view.h"
+#include "word_counts.h"
 #include "word_tree.h"
 
 #include <QAction>
@@ -46,12 +50,16 @@
 #include <QToolButton>
 #include <QVBoxLayout>
 
+#include <ctime>
+
 //-----------------------------------------------------------------------------
 
 Board::Board(QWidget* parent)
 : QWidget(parent), m_paused(false), m_wrong(false), m_valid(true), m_score_type(1), m_size(0), m_minimum(0), m_maximum(0), m_max_score(0), m_generator(0) {
 	m_generator = new Generator(this);
 	connect(m_generator, SIGNAL(finished()), this, SLOT(gameStarted()));
+	connect(m_generator, SIGNAL(optimizingStarted()), this, SIGNAL(optimizingStarted()));
+	connect(m_generator, SIGNAL(optimizingFinished()), this, SIGNAL(optimizingFinished()));
 
 	m_view = new View(0, this);
 
@@ -64,7 +72,7 @@ Board::Board(QWidget* parent)
 	m_max_score_details = new QToolButton(this);
 	m_max_score_details->setAutoRaise(true);
 	m_max_score_details->setIconSize(QSize(16,16));
-	m_max_score_details->setIcon(QIcon(":/info.png"));
+	m_max_score_details->setIcon(QIcon::fromTheme("dialog-information", QIcon(":/dialog-information.png")));
 	m_max_score_details->setToolTip(tr("Details"));
 	connect(m_max_score_details, SIGNAL(clicked()), this, SLOT(showMaximumWords()));
 
@@ -82,18 +90,30 @@ Board::Board(QWidget* parent)
 	connect(m_guess, SIGNAL(returnPressed()), this, SLOT(guess()));
 	connect(m_view, SIGNAL(mousePressed()), m_guess, SLOT(setFocus()));
 
+	int size = style()->pixelMetric(QStyle::PM_ToolBarIconSize);
+
 	m_clear_button = new QToolButton(this);
 	m_clear_button->setAutoRaise(true);
-	m_clear_button->setIconSize(QSize(24,24));
-	m_clear_button->setIcon(QIcon(":/edit-clear.png"));
+	m_clear_button->setIconSize(QSize(size,size));
+	QIcon clear_fallback(":/tango/32x32/actions/edit-clear.png");
+	clear_fallback.addFile(":/tango/24x24/actions/edit-clear.png");
+	clear_fallback.addFile(":/tango/22x22/actions/edit-clear.png");
+	clear_fallback.addFile(":/tango/16x16/actions/edit-clear.png");
+	m_clear_button->setIcon(QIcon::fromTheme("edit-clear", clear_fallback));
 	m_clear_button->setToolTip(tr("Clear"));
+	m_clear_button->setEnabled(false);
 	connect(m_clear_button, SIGNAL(clicked()), this, SLOT(clearGuess()));
 
 	m_guess_button = new QToolButton(this);
 	m_guess_button->setAutoRaise(true);
-	m_guess_button->setIconSize(QSize(24,24));
-	m_guess_button->setIcon(QIcon(":/list-add.png"));
+	m_guess_button->setIconSize(QSize(size,size));
+	QIcon guess_fallback(":/tango/32x32/actions/list-add.png");
+	guess_fallback.addFile(":/tango/24x24/actions/list-add.png");
+	guess_fallback.addFile(":/tango/22x22/actions/list-add.png");
+	guess_fallback.addFile(":/tango/16x16/actions/list-add.png");
+	m_guess_button->setIcon(QIcon::fromTheme("list-add", guess_fallback));
 	m_guess_button->setToolTip(tr("Guess"));
+	m_guess_button->setEnabled(false);
 	connect(m_guess_button, SIGNAL(clicked()), this, SLOT(guess()));
 
 	QHBoxLayout* guess_layout = new QHBoxLayout;
@@ -128,6 +148,9 @@ Board::Board(QWidget* parent)
 	int width = guess_layout->sizeHint().width();
 	m_tabs->setFixedWidth(width);
 
+	m_counts = new WordCounts(this);
+	m_counts->setMinimumWidth(width);
+
 	// Lay out board
 	QGridLayout* layout = new QGridLayout(this);
 	layout->setColumnStretch(1, 1);
@@ -137,6 +160,42 @@ Board::Board(QWidget* parent)
 	layout->addWidget(m_clock, 0, 1, Qt::AlignCenter);
 	layout->addWidget(m_view, 1, 1);
 	layout->addLayout(score_layout, 2, 1, Qt::AlignCenter);
+	layout->addWidget(m_counts, 3, 0, 1, 2);
+}
+
+//-----------------------------------------------------------------------------
+
+Board::~Board() {
+	QSettings game;
+	if (isFinished()) {
+		// Clear current game
+		game.remove("Current");
+	} else {
+		// Save current game
+		game.beginGroup("Current");
+
+		QStringList found;
+		for (int i = 0; i < m_found->topLevelItemCount(); ++i) {
+			found += m_found->topLevelItem(i)->text(2);
+		}
+		game.setValue("Found", found);
+
+		QVariantList positions;
+		QString word;
+		foreach (const QPoint& position, m_positions) {
+			positions.append(position);
+			word.append(m_cells[position.x()][position.y()]->text().toUpper());
+		}
+		if (!m_wrong && (m_guess->text() == word)) {
+			game.setValue("Guess", m_guess->text());
+			game.setValue("GuessPositions", positions);
+		} else {
+			game.remove("Guess");
+			game.remove("GuessPositions");
+		}
+
+		m_clock->save(game);
+	}
 }
 
 //-----------------------------------------------------------------------------
@@ -154,9 +213,51 @@ void Board::abort() {
 
 //-----------------------------------------------------------------------------
 
-void Board::generate(bool higher_scores, int size, int timer, unsigned int seed) {
+void Board::generate(const QSettings& game) {
+	// Find values
+	int size = qBound(4, game.value("Size").toInt(), 5);
+	int density = qBound(0, game.value("Density").toInt(), 3);
+	int minimum = game.value("Minimum").toInt();
+	if (size == 4) {
+		minimum = qBound(3, minimum, 6);
+	} else {
+		minimum = qBound(4, minimum, 7);
+	}
+	int timer = qBound(0, game.value("TimerMode").toInt(), Clock::TotalTimers - 1);
+	QStringList letters = game.value("Letters").toStringList();
+	unsigned int seed = Random(time(0)).nextInt();
+
+	LanguageSettings settings;
+	int language = game.value("Language", settings.language()).toInt();
+	QString dice = game.value("Dice", settings.dice()).toString();
+	QString words = game.value("Words", settings.words()).toString();
+	QString dictionary = game.value("Dictionary", settings.dictionary()).toString();
+
+	bool is_hebrew = (language == QLocale::Hebrew);
+	m_found->setHebrew(is_hebrew);
+	m_missed->setHebrew(is_hebrew);
+
+	// Store values
+	{
+		QSettings settings;
+		settings.beginGroup("Current");
+		settings.setValue("Version", 2);
+		settings.setValue("Size", size);
+		settings.setValue("Density", density);
+		settings.setValue("Minimum", minimum);
+		settings.setValue("TimerMode", timer);
+		settings.setValue("Language", language);
+		settings.setValue("Dice", dice);
+		settings.setValue("Words", words);
+		settings.setValue("Dictionary", dictionary);
+		if (!letters.isEmpty()) {
+			settings.setValue("Letters", letters);
+		}
+	}
+
+	// Create new game
 	m_generator->cancel();
-	m_generator->create(higher_scores, size, timer, seed);
+	m_generator->create(density, size, minimum, timer, letters, seed);
 }
 
 //-----------------------------------------------------------------------------
@@ -183,12 +284,6 @@ QString Board::sizeToString(int size) {
 
 //-----------------------------------------------------------------------------
 
-void Board::setHigherScoringBoards(bool higher) {
-	QSettings().setValue("Board/HigherScores", higher);
-}
-
-//-----------------------------------------------------------------------------
-
 void Board::setShowMaximumScore(QAction* show) {
 	int score_type = show->data().toInt();
 	QSettings().setValue("ShowMaximumScore", score_type);
@@ -216,24 +311,44 @@ void Board::setShowMissedWords(bool show) {
 
 //-----------------------------------------------------------------------------
 
+void Board::setShowWordCounts(bool show) {
+	QSettings().setValue("ShowWordCounts", show);
+	m_counts->setVisible(show);
+}
+
+//-----------------------------------------------------------------------------
+
+void Board::updateScoreColor() {
+	updateScore();
+}
+
+//-----------------------------------------------------------------------------
+
 void Board::gameStarted() {
 	// Load settings
+	QSettings settings;
+	settings.beginGroup("Current");
+
 	m_clock->setTimer(m_generator->timer());
 	if (m_generator->size() != m_size) {
 		m_size = m_generator->size();
 		m_cells = QVector<QVector<Letter*> >(m_size, QVector<Letter*>(m_size));
-		m_minimum = m_size - 1;
 		m_maximum = m_size * m_size;
 		m_guess->setMaxLength(m_maximum);
 	}
+	m_minimum = m_generator->minimum();
 	m_max_score = m_generator->maxScore();
 	m_max_score_details->hide();
 	m_letters = m_generator->letters();
 	m_solutions = m_generator->solutions();
+	m_counts->setWords(m_solutions.keys());
 	m_found->setDictionary(m_generator->dictionary());
+	m_found->setTrie(m_generator->trie());
 	m_missed->setDictionary(m_generator->dictionary());
+	m_missed->setTrie(m_generator->trie());
+	settings.setValue("Letters", m_letters);
 
-	// Create cells
+	// Create board
 	QFont f = font();
 	f.setBold(true);
 	f.setPointSize(20);
@@ -254,10 +369,11 @@ void Board::gameStarted() {
 	m_view->setMinimumSize(board_size + 4, board_size + 4);
 	m_view->fitInView(m_view->sceneRect(), Qt::KeepAspectRatio);
 
-	QPainterPath path;
-	path.addRoundedRect(0, 0, board_size, board_size, 5, 5);
-	scene->addPath(path, Qt::NoPen, QColor("#0057ae"));
+	BeveledRect* rect = new BeveledRect(board_size);
+	rect->setColor("#0057ae");
+	scene->addItem(rect);
 
+	// Create cells
 	for (int r = 0; r < m_size; ++r) {
 		for (int c = 0; c < m_size; ++c) {
 			Letter* cell = new Letter(f, cell_size, QPoint(c,r));
@@ -274,6 +390,7 @@ void Board::gameStarted() {
 	m_tabs->setTabEnabled(1, false);
 
 	// Clear previous words
+	m_paused = false;
 	emit pauseAvailable(true);
 	m_clear_button->setEnabled(true);
 	m_guess_button->setEnabled(true);
@@ -293,6 +410,24 @@ void Board::gameStarted() {
 		m_missed->addWord(solution);
 	}
 
+	// Add found words
+	QStringList found = settings.value("Found").toStringList();
+	foreach (const QString& text, found) {
+		QTreeWidgetItem* item = m_found->findItems(text, Qt::MatchExactly, 2).value(0);
+		if (m_missed->findItems(text, Qt::MatchExactly, 2).value(0) && !item) {
+			item = m_found->addWord(text);
+			delete m_missed->findItems(text, Qt::MatchExactly, 2).first();
+			m_counts->findWord(text);
+		}
+	}
+
+	// Add guess
+	m_guess->setText(settings.value("Guess").toString());
+	QVariantList positions = settings.value("GuessPositions").toList();
+	foreach (const QVariant& position, positions) {
+		m_positions.append(position.toPoint());
+	}
+
 	// Show errors
 	QString error = m_generator->error();
 	if (!error.isEmpty()) {
@@ -305,8 +440,13 @@ void Board::gameStarted() {
 	emit started();
 	if (m_missed->topLevelItemCount() > 0) {
 		m_clock->start();
+		if (settings.contains("TimerDetails/Time")) {
+			m_clock->load(settings);
+		}
 		updateScore();
 		updateClickableStatus();
+		updateButtons();
+		highlightWord();
 	} else {
 		m_clock->stop();
 	}
@@ -324,6 +464,7 @@ void Board::clearGuess() {
 	m_found->clearSelection();
 	m_missed->clearSelection();
 	m_guess->setFocus();
+	updateButtons();
 }
 
 //-----------------------------------------------------------------------------
@@ -331,21 +472,22 @@ void Board::clearGuess() {
 void Board::guess() {
 	if (!isFinished() && !m_paused) {
 		QString text = m_guess->text().trimmed().toUpper();
-		if (text.isEmpty() || text.length() < m_minimum || text.length() > m_maximum || !m_valid || m_positions.isEmpty()) {
+		if (text.isEmpty() || text.length() < m_minimum || text.length() > m_maximum || !m_valid || m_positions.isEmpty() || m_wrong) {
 			return;
 		}
 		if (!m_solutions.contains(text)) {
 			m_wrong = true;
 			highlightWord();
+			updateButtons();
 			m_clock->addIncorrectWord(Solver::score(text));
 			return;
 		}
 
 		// Create found item
-		QTreeWidgetItem* item = m_found->findItems(text, Qt::MatchExactly).value(0);
+		QTreeWidgetItem* item = m_found->findItems(text, Qt::MatchExactly, 2).value(0);
 		if (item == 0) {
 			item = m_found->addWord(text);
-			delete m_missed->findItems(text, Qt::MatchExactly, 0).first();
+			delete m_missed->findItems(item->text(2), Qt::MatchExactly, 2).first();
 
 			m_clock->addWord(item->data(0, Qt::UserRole).toInt());
 			updateScore();
@@ -357,6 +499,8 @@ void Board::guess() {
 			} else {
 				solutions.prepend(m_positions);
 			}
+
+			m_counts->findWord(text);
 		}
 		m_found->scrollToItem(item, QAbstractItemView::PositionAtCenter);
 		m_found->clearSelection();
@@ -392,7 +536,10 @@ void Board::guessChanged() {
 		m_guess->setText(word);
 		m_guess->setCursorPosition(pos);
 
-		QList<QList<QPoint> > solutions = m_solutions.value(word, Solver(word, m_letters, 0).solutions().value(word));
+		Trie trie(word);
+		Solver solver(trie, m_size, 0);
+		solver.solve(m_letters);
+		QList<QList<QPoint> > solutions = m_solutions.value(word, solver.solutions().value(word));
 		m_valid = !solutions.isEmpty();
 		if (m_valid) {
 			int index = 0;
@@ -444,6 +591,8 @@ void Board::guessChanged() {
 		m_positions.clear();
 		updateClickableStatus();
 	}
+
+	updateButtons();
 }
 
 //-----------------------------------------------------------------------------
@@ -479,7 +628,7 @@ void Board::wordSelected() {
 		return;
 	}
 
-	QString word = items.first()->text(0);
+	QString word = items.first()->text(2);
 	if (!word.isEmpty() && word != m_guess->text()) {
 		m_guess->setText(word);
 		m_positions = m_solutions.value(word).value(0);
@@ -487,12 +636,13 @@ void Board::wordSelected() {
 		updateClickableStatus();
 		highlightWord();
 	}
+
+	updateButtons();
 }
 
 //-----------------------------------------------------------------------------
 
 void Board::letterClicked(Letter* letter) {
-	m_wrong = false;
 
 	// Handle adding a letter to the guess
 	if (!m_positions.contains(letter->position())) {
@@ -500,6 +650,7 @@ void Board::letterClicked(Letter* letter) {
 		word.append(letter->text().toUpper());
 		m_guess->setText(word);
 
+		m_wrong = false;
 		m_positions.append(letter->position());
 		clearHighlight();
 		updateClickableStatus();
@@ -508,13 +659,14 @@ void Board::letterClicked(Letter* letter) {
 		selectGuess();
 	// Handle making or clearing a guess
 	} else if (letter->position() == m_positions.last()) {
-		if (m_positions.count() >= 3) {
-			guess();
-		} else if (m_positions.count() == 1) {
+		if (m_positions.count() == 1) {
 			clearGuess();
+		} else {
+			guess();
 		}
 	// Handle backing up in a guess
 	} else {
+		m_wrong = false;
 		m_guess->clear();
 		m_positions = m_positions.mid(0, m_positions.indexOf(letter->position()) + 1);
 		clearHighlight();
@@ -529,6 +681,8 @@ void Board::letterClicked(Letter* letter) {
 
 		selectGuess();
 	}
+
+	updateButtons();
 }
 
 //-----------------------------------------------------------------------------
@@ -536,10 +690,10 @@ void Board::letterClicked(Letter* letter) {
 void Board::highlightWord(const QList<QPoint>& positions, const QColor& color) {
 	Q_ASSERT(!positions.isEmpty());
 
-	m_cells[positions.at(0).x()][positions.at(0).y()]->setBrush(color);
+	m_cells[positions.at(0).x()][positions.at(0).y()]->setColor(color);
 	for (int i = 1; i < positions.count(); ++i) {
 		const QPoint& pos1 = positions.at(i);
-		m_cells[pos1.x()][pos1.y()]->setBrush(color);
+		m_cells[pos1.x()][pos1.y()]->setColor(color);
 
 		const QPoint& pos0 = m_positions.at(i - 1);
 		QLineF line(pos0, pos1);
@@ -550,7 +704,7 @@ void Board::highlightWord(const QList<QPoint>& positions, const QColor& color) {
 	QColor border = Qt::white;
 	for (int i = positions.count() - 1; i >= 0; --i) {
 		const QPoint& position = positions.at(i);
-		m_cells[position.x()][position.y()]->setPen(QPen(border, 2));
+		m_cells[position.x()][position.y()]->setCellColor(border);
 		border.setAlpha(border.alpha() - alpha);
 	}
 }
@@ -590,11 +744,11 @@ void Board::highlightWord() {
 //-----------------------------------------------------------------------------
 
 void Board::clearHighlight() {
-	QColor color = !isFinished() ? Qt::white : QColor("#dddddd");
+	QColor color = !isFinished() ? Qt::white : QColor("#aaa");
 	for (int c = 0; c < m_size; ++c) {
 		for (int r = 0; r < m_size; ++r) {
-			m_cells[c][r]->setBrush(color);
-			m_cells[c][r]->setPen(Qt::NoPen);
+			m_cells[c][r]->setColor(color);
+			m_cells[c][r]->setCellColor(QColor());
 			m_cells[c][r]->setArrow(-1, 0);
 		}
 	}
@@ -622,9 +776,21 @@ int Board::updateScore() {
 	}
 
 	if (m_score_type == 2 || (m_score_type == 1 && isFinished())) {
-		m_score->setText(tr("%1 of %n point(s)", "", m_max_score).arg(score));
+		if (score > 3) {
+			m_score->setText(tr("%1 of %n point(s)", "", m_max_score).arg(score));
+		} else if (score == 3) {
+			m_score->setText(tr("3 of %n point(s)", "", m_max_score));
+		} else if (score == 2) {
+			m_score->setText(tr("2 of %n point(s)", "", m_max_score));
+		} else if (score == 1) {
+			m_score->setText(tr("1 of %n point(s)", "", m_max_score));
+		} else {
+			m_score->setText(tr("0 of %n point(s)", "", m_max_score));
+		}
+		m_counts->setMaximumsVisible(true);
 	} else {
 		m_score->setText(tr("%n point(s)", "", score));
+		m_counts->setMaximumsVisible(false);
 	}
 
 	QFont f = font();
@@ -679,6 +845,15 @@ void Board::updateClickableStatus() {
 
 //-----------------------------------------------------------------------------
 
+void Board::updateButtons() {
+	QString text = m_guess->text();
+	bool has_guess = !text.isEmpty();
+	m_clear_button->setEnabled(has_guess);
+	m_guess_button->setEnabled(has_guess && text.length() >= m_minimum && text.length() <= m_maximum && m_valid && !m_positions.isEmpty() && !m_wrong);
+}
+
+//-----------------------------------------------------------------------------
+
 void Board::showMaximumWords() {
 	QDialog dialog(window(), Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint);
 	dialog.setWindowTitle(tr("Details"));
